// ----------------------------------------------
#ifndef __CGPUMESH__
#define __CGPUMESH__
// ----------------------------------------------
#include <vector>
#include <Typelist.h>
#include <static_check.h>
// ----------------------------------------------
#include "VertexFormat.h"
// ----------------------------------------------
namespace GPUmesh {
// ----------------------------------------------
template <template <class,class> class T_CompilePolicy,
		      class                        T_VertexFormat,
          class                        T_IndexType=Loki::NullType>
class GPUmesh_generic
{
public:

  typedef T_CompilePolicy<T_VertexFormat,T_IndexType>  compile_policy;
  typedef T_VertexFormat                               vertex_format;
  typedef typename compile_policy::vertex_data         vertex_data;
  typedef typename compile_policy::vertex_format_desc  vertex_format_desc;

protected:

  typename compile_policy::mesh_descriptor m_Mesh;
  typename compile_policy::vertex_storage  m_Storage;
  vertex_data                              m_CurrentVertex;
  int                                      m_iPrimType;
  bool                                     m_bEmpty;
  typename compile_policy::index_storage   m_Indices;

  int push_current()
  {
      m_Storage.push_vertex(m_CurrentVertex);
      return (m_Storage.nbVertices()-1);
  }

  void clear()
  {
    for (int i=0;i<MVF_sizeof<T_VertexFormat>::value;i++)
      m_CurrentVertex.data[i]=0;
  }

public:

  GPUmesh_generic() : m_bEmpty(true) 
    {
      // check if the vertex format defines a position !
      STATIC_CHECK(FormatDescriptor<T_VertexFormat>::has_vertex,vertex_format_should_contain_a_vertex);
      
      clear();
    }

  ~GPUmesh_generic()
  {
    if (!m_bEmpty)
      compile_policy::free(m_Mesh);
  }

  void begin(int ptype)
  {
    if (!m_bEmpty)
      compile_policy::free(m_Mesh);
    clear();
    m_Storage.clear();
    m_Indices.clear();
    m_iPrimType=ptype;
  }

  void end()
  {
    compile_policy::compile(m_iPrimType,m_Storage,m_Indices,m_Mesh);
  }

  void draw()
  {
    compile_policy::draw(m_Mesh);
  }

  void bind()
  {
    compile_policy::bind(m_Mesh);
  }

  void unbind()
  {
    compile_policy::unbind(m_Mesh);
  }

  void     render()
  {
    bind();
    draw();
    unbind();
  }

  // vertex methods

  int vertex_2(typename vertex_format_desc::type_vertex::type x,
                typename vertex_format_desc::type_vertex::type y)
  {
    if (!vertex_format_desc::has_vertex)
      return (-1);
    STATIC_CHECK(vertex_format_desc::type_vertex::components == 2,vertex_format_does_not_support_vertex2);
    typename vertex_format_desc::type_vertex t;
    t.data[0]=x;
    t.data[1]=y;
    memcpy(&(m_CurrentVertex.data[vertex_format_desc::offset_vertex]),
           t.data,
           vertex_format_desc::type_vertex::size_of);

    return (push_current());
  }

  int vertex_3(typename vertex_format_desc::type_vertex::type x,
                typename vertex_format_desc::type_vertex::type y,
                typename vertex_format_desc::type_vertex::type z)
  {
    if (!vertex_format_desc::has_vertex)
      return (-1);
    STATIC_CHECK(vertex_format_desc::type_vertex::components == 3,vertex_format_does_not_support_vertex3);
    typename vertex_format_desc::type_vertex t;
    t.data[0]=x;
    t.data[1]=y;
    t.data[2]=z;
    int test0=vertex_format_desc::offset_vertex;
    int test1=vertex_format_desc::type_vertex::size_of;
    memcpy(&(m_CurrentVertex.data[vertex_format_desc::offset_vertex]),
           t.data,
           vertex_format_desc::type_vertex::size_of);

    return (push_current());
  }

  int vertex_4(typename vertex_format_desc::type_vertex::type x,
                typename vertex_format_desc::type_vertex::type y,
                typename vertex_format_desc::type_vertex::type z,
                typename vertex_format_desc::type_vertex::type w)
  {
    if (!vertex_format_desc::has_vertex)
      return (-1);
    STATIC_CHECK(vertex_format_desc::type_vertex::components == 4,vertex_format_does_not_support_vertex4);
    typename vertex_format_desc::type_vertex t;
    t.data[0]=x;
    t.data[1]=y;
    t.data[2]=z;
    t.data[3]=z;
    memcpy(&(m_CurrentVertex.data[vertex_format_desc::offset_vertex]),
           t.data,
           vertex_format_desc::type_vertex::size_of);

    return (push_current());
  }

  // define some ugly macros to ease method creation for attributes

#define DATA_METHOD_1(func_name,name) void func_name(typename vertex_format_desc::type_##name::type x) { \
    if (!vertex_format_desc::has_##name) \
      return; \
    if (vertex_format_desc::type_##name::components != 1) \
      return; \
    typename vertex_format_desc::type_##name t; \
    t.data[0]=x; \
    memcpy(&(m_CurrentVertex.data[vertex_format_desc::offset_##name]), \
           t.data, \
           vertex_format_desc::type_##name::size_of); } \
           \
    void func_name(MVF_empty x) {}

#define DATA_METHOD_2(func_name,name) void func_name(typename vertex_format_desc::type_##name::type x,   \
                                                     typename vertex_format_desc::type_##name::type y) { \
    if (!vertex_format_desc::has_##name) \
      return; \
    if (vertex_format_desc::type_##name::components != 2) \
      return; \
    typename vertex_format_desc::type_##name t; \
    t.data[0]=x; \
    t.data[1]=y; \
    memcpy(&(m_CurrentVertex.data[vertex_format_desc::offset_##name]), \
           t.data, \
           vertex_format_desc::type_##name::size_of); } \
           \
    void func_name(MVF_empty x,MVF_empty y) {}

#define DATA_METHOD_3(func_name,name) void func_name(typename vertex_format_desc::type_##name::type x,   \
                                                     typename vertex_format_desc::type_##name::type y,   \
                                                     typename vertex_format_desc::type_##name::type z) { \
    if (!vertex_format_desc::has_##name) \
      return; \
    if (vertex_format_desc::type_##name::components != 3) \
      return; \
    typename vertex_format_desc::type_##name t; \
    t.data[0]=x; \
    t.data[1]=y; \
    t.data[2]=z; \
    memcpy(&(m_CurrentVertex.data[vertex_format_desc::offset_##name]), \
           t.data, \
           vertex_format_desc::type_##name::size_of); } \
           \
   void func_name(MVF_empty x,MVF_empty y,MVF_empty z) {}

#define DATA_METHOD_4(func_name,name) void func_name(typename vertex_format_desc::type_##name::type x,   \
                                                     typename vertex_format_desc::type_##name::type y,   \
                                                     typename vertex_format_desc::type_##name::type z,   \
                                                     typename vertex_format_desc::type_##name::type w) { \
    if (!vertex_format_desc::has_##name) \
      return; \
    if (vertex_format_desc::type_##name::components != 4) \
      return; \
    typename vertex_format_desc::type_##name t; \
    t.data[0]=x; \
    t.data[1]=y; \
    t.data[2]=z; \
    t.data[3]=w; \
    memcpy(&(m_CurrentVertex.data[vertex_format_desc::offset_##name]), \
           t.data, \
           vertex_format_desc::type_##name::size_of); } \
           \
    void func_name(MVF_empty x,MVF_empty y,MVF_empty z,MVF_empty w) {}

  // use macros to define attribute methods

  DATA_METHOD_3(normal,normal)
  DATA_METHOD_3(color0_3,color0)
  DATA_METHOD_4(color0_4,color0)
  DATA_METHOD_3(color1_3,color1)
  DATA_METHOD_4(color1_4,color1)
  DATA_METHOD_1(texcoord0_1,texcoord0)
  DATA_METHOD_2(texcoord0_2,texcoord0)
  DATA_METHOD_3(texcoord0_3,texcoord0)
  DATA_METHOD_4(texcoord0_4,texcoord0)
  DATA_METHOD_1(texcoord1_1,texcoord1)
  DATA_METHOD_2(texcoord1_2,texcoord1)
  DATA_METHOD_3(texcoord1_3,texcoord1)
  DATA_METHOD_4(texcoord1_4,texcoord1)
  DATA_METHOD_1(texcoord2_1,texcoord2)
  DATA_METHOD_2(texcoord2_2,texcoord2)
  DATA_METHOD_3(texcoord2_3,texcoord2)
  DATA_METHOD_4(texcoord2_4,texcoord2)
  DATA_METHOD_1(texcoord3_1,texcoord3)
  DATA_METHOD_2(texcoord3_2,texcoord3)
  DATA_METHOD_3(texcoord3_3,texcoord3)
  DATA_METHOD_4(texcoord3_4,texcoord3)
  DATA_METHOD_1(texcoord4_1,texcoord4)
  DATA_METHOD_2(texcoord4_2,texcoord4)
  DATA_METHOD_3(texcoord4_3,texcoord4)
  DATA_METHOD_4(texcoord4_4,texcoord4)
  DATA_METHOD_1(texcoord5_1,texcoord5)
  DATA_METHOD_2(texcoord5_2,texcoord5)
  DATA_METHOD_3(texcoord5_3,texcoord5)
  DATA_METHOD_4(texcoord5_4,texcoord5)
  DATA_METHOD_1(texcoord6_1,texcoord6)
  DATA_METHOD_2(texcoord6_2,texcoord6)
  DATA_METHOD_3(texcoord6_3,texcoord6)
  DATA_METHOD_4(texcoord6_4,texcoord6)
  DATA_METHOD_1(texcoord7_1,texcoord7)
  DATA_METHOD_2(texcoord7_2,texcoord7)
  DATA_METHOD_3(texcoord7_3,texcoord7)
  DATA_METHOD_4(texcoord7_4,texcoord7)

  // index methods

  void index(T_IndexType idx)
  {
    if (Loki::IsSameType<T_IndexType,Loki::NullType>::value)
      return;
    m_Indices.push_back(idx);
  }

};
// ----------------------------------------------
} // namespace GPUmesh
// ----------------------------------------------
#endif
// ----------------------------------------------
